// File_Avc - Info for AVC Video files
// Copyright (C) 2006-2007 Jerome Martinez, Zen@MediaArea.net
//
// This library is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library. If not, see <http://www.gnu.org/licenses/>.
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
// Compilation conditions
#include <MediaInfo/Setup.h>
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#if defined(MEDIAINFO_AVC_YES)
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "MediaInfo/Video/File_Avc.h"
#include "math.h"
using namespace ZenLib;
//---------------------------------------------------------------------------

namespace MediaInfoLib
{

//***************************************************************************
// Constants
//***************************************************************************

//---------------------------------------------------------------------------
const size_t Avc_Errors_MaxCount=32;

const int8u Avc_PixelAspectRatio_Size=14;
const float32 Avc_PixelAspectRatio[]=
{
    (float32)1, //Reserved
    (float32)1,
    (float32)12/(float32)11,
    (float32)10/(float32)11,
    (float32)16/(float32)11,
    (float32)40/(float32)33,
    (float32)24/(float32)11,
    (float32)20/(float32)11,
    (float32)32/(float32)11,
    (float32)80/(float32)33,
    (float32)18/(float32)11,
    (float32)15/(float32)11,
    (float32)64/(float32)33,
    (float32)160/(float32)99,
};

const char* Avc_primary_pic_type[]=
{
    "I",
    "I, P",
    "I, P, B",
    "SI",
    "SI, SP",
    "I, SI",
    "I, SI, P, SP",
    "I, SI, P, SP, B",
};

const char* Avc_slice_type[]=
{
    "P",
    "B",
    "I",
    "SP",
    "SI",
    "P",
    "B",
    "I",
    "SP",
    "SI",
};
//---------------------------------------------------------------------------

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

//---------------------------------------------------------------------------
File_Avc::File_Avc()
:File__Analyze()
{
    //In
    Frame_Count_Valid=2;
    FrameIsAlwaysComplete=false;

    //Count of a Packets
    Frame_Count=0;
    frame_num_LastOne=(int32u)-1;

    //From seq_parameter_set
    pic_width_in_mbs_minus1=0;
    pic_height_in_map_units_minus1=0;
    log2_max_frame_num_minus4=0;
    profile_idc=0;
    level_idc=0;
    aspect_ratio_idc=0xFF;
    sar_width=0;
    sar_height=0;
    frame_mbs_only_flag=false;

    //Default values
    Stream[0x07].Searching_Payload=true; //seq_parameter_set
    Stream[0x09].Searching_Payload=true; //access_unit_delimiter
    for (int8u Pos=0xB9; Pos!=0x00; Pos++)
        Stream[Pos].Searching_Payload=true; //Testing MPEG-PS
}

//***************************************************************************
// Buffer
//***************************************************************************

//---------------------------------------------------------------------------
bool File_Avc::Header_Begin()
{
    //Trailing 0x00
    if (Synched)
    {
        while(Buffer_Offset+3<Buffer_Size && CC1(Buffer+Buffer_Offset)==0x00 && CC3(Buffer+Buffer_Offset)!=0x000001)
            Buffer_Offset++;
    }

    //Must have enough buffer for having header
    if (Buffer_Offset+4>Buffer_Size)
        return false;

    //Quick test of synchro
    if (Synched && CC3(Buffer+Buffer_Offset)!=0x000001)
    {
        Trusted_IsNot("AVC, Synchronisation lost");
        Synched=false;
    }

    //Synchro
    if (!Synched && !Synchronize())
        return false;

    //Quick search
    if (!Header_Parser_QuickSearch())
        return false;

    return true;
}

//---------------------------------------------------------------------------
void File_Avc::Header_Parse()
{
    //Parsing
    int8u nal_unit_type;
    Skip_B3(                                                    "sync");
    BS_Begin();
    Mark_0 ();
    Skip_S1( 2,                                                 "nal_ref_idc");
    Get_S1 ( 5, nal_unit_type,                                  "nal_unit_type");
    BS_End();
    if (!Header_Parse_Fill_Size())
    {
        Element_WaitForMoreData();
        return;
    }

    //Filling
    Header_Fill_Code(nal_unit_type, Ztring().From_CC1(nal_unit_type));
}

//---------------------------------------------------------------------------
bool File_Avc::Header_Parse_Fill_Size()
{
    //Look for next Sync word
    if (Buffer_Offset_Temp==0) //Buffer_Offset_Temp is not 0 if Header_Parse_Fill_Size() has already parsed first frames
        Buffer_Offset_Temp=Buffer_Offset+4;
    while (Buffer_Offset_Temp+4<=Buffer_Size
        && CC3(Buffer+Buffer_Offset_Temp)!=0x000001)
        Buffer_Offset_Temp++;

    //Must wait more data?
    if (Buffer_Offset_Temp+4>Buffer_Size)
    {
        if (FrameIsAlwaysComplete || File_Offset+Buffer_Size==File_Size)
            Buffer_Offset_Temp=Buffer_Size; //We are sure that the next bytes are a start
        else
            return false;
    }

    //OK, we continue
    Header_Fill_Size(Buffer_Offset_Temp-Buffer_Offset);
    Buffer_Offset_Temp=0;
    return true;
}

//---------------------------------------------------------------------------
bool File_Avc::Data_Parse()
{
    /*
    //AVC buffer
    Buffer_ToSave=Buffer;
    Buffer_Size_ToSave=Buffer_Size;

    //Searching emulation_prevention_three_byte
    size_t ThreeByte_Buffer_Offset=0;
    std::vector<size_t> ThreeByte_List;
    while (Buffer_Offset_3Bytes+3<=Buffer_Size)
    {
        if (CC3(Buffer+ThreeByte_Buffer_Offset)==0x000003)
            ThreeByte_List.push_back(ThreeByte_Buffer_Offset);
        Buffer_Offset_3Bytes++;
    }
    if (ThreeByte_List.empty())
    {
        //We must change the buffer for keeping out
    }
    */


    //Parsing
    switch (Element_Code)
    {
        case 0x01 : slice_layer_without_partitioning_non_IDR(); break;
        case 0x02 : Element_Name("slice_data_partition_a_layer"); break;
        case 0x03 : Element_Name("slice_data_partition_b_layer"); break;
        case 0x04 : Element_Name("slice_data_partition_c_layer"); break;
        case 0x05 : slice_layer_without_partitioning_IDR(); break;
        case 0x06 : sei(); break;
        case 0x07 : seq_parameter_set(); break;
        case 0x08 : pic_parameter_set(); break;
        case 0x09 : access_unit_delimiter(); break;
        case 0x10 : Element_Name("end_of_seq"); break;
        case 0x11 : Element_Name("end_of_stream"); break;
        case 0x12 : Element_Name("filler_data"); break;
        default :
            if (Element_Code<=0x23)
                Element_Name("reserved");
            else
                Trusted_IsNot("Unattended element!");
    }

    return true;
}

//***************************************************************************
// Elements
//***************************************************************************

//---------------------------------------------------------------------------
// Packet "01"
void File_Avc::slice_layer_without_partitioning_non_IDR()
{
    Element_Name("slice_layer_without_partitioning (non-IDR)");

    //Parsing
    slice_header();
}

//---------------------------------------------------------------------------
// Packet "05"
void File_Avc::slice_layer_without_partitioning_IDR()
{
    Element_Name("slice_layer_without_partitioning (IDR)");

    //Parsing
    slice_header();

    FILLING_BEGIN();
        //NextCode
        for (int8u Pos=0x01; Pos<=0x05; Pos++)
            NextCode_Add(Pos);
    FILLING_END();
}

//---------------------------------------------------------------------------
//
void File_Avc::slice_header()
{
    //Parsing
    int32u slice_type, frame_num;
    BS_Begin();
    Skip_UE(                                                    "first_mb_in_slice");
    Get_UE (slice_type,                                         "slice_type"); if (slice_type<9) Param_Info(Avc_slice_type[slice_type]);
    Skip_UE(                                                    "pic_parameter_set_id");
    Get_BS (log2_max_frame_num_minus4, frame_num,               "frame_num");
    if (!frame_mbs_only_flag)
    {
        TEST_SB_SKIP(                                           "field_pic_flag");
            Skip_SB(                                            "bottom_field_flag");
        TEST_SB_END();
    }
    //TODO...

    FILLING_BEGIN();
        //Counting
        if (File_Offset+Buffer_Offset+Element_Size==File_Size)
            Frame_Count_Valid=Frame_Count; //Finalize frames in case of there are less than Frame_Count_Valid frames
        if (frame_num!=frame_num_LastOne)
        {
            Frame_Count++;
            frame_num_LastOne=frame_num;
        }

        //Name
        Element_Info(Ztring::ToZtring(Frame_Count));

        //Filling only if not already done
        if (Frame_Count>=Frame_Count_Valid && Count_Get(Stream_Video)==0)
            slice_header_Fill();
    FILLING_END();
}

//---------------------------------------------------------------------------
//
void File_Avc::slice_header_Fill()
{
    //Calculating - Pixels
    int32u Width=(pic_width_in_mbs_minus1+1)*16;
    int32u Height=(pic_height_in_map_units_minus1+1)*(2-frame_mbs_only_flag)*16;

    //Calculating - Profile
    Ztring ProfileS;
    switch (profile_idc)
    {
        case 0x42 : ProfileS=_T("Baseline"); break;
        case 0x4D : ProfileS=_T("Main"); break;
        case 0x58 : ProfileS=_T("Extended"); break;
        case 0x64 : ProfileS=_T("High"); break;
        case 0x6E : ProfileS=_T("High 10"); break;
        case 0x7A : ProfileS=_T("High 4:2:2"); break;
        case 0x90 : ProfileS=_T("High 4:4:4"); break;
        default   : ProfileS.From_Number(profile_idc); break;
    }

    //Calculating - Level
    Ztring LevelS;
    LevelS.From_Number(((float)level_idc)/10, 1); //Level is Value*10, can have one digit example 5.1

    //Calculating - PixelAspectRatio
    float32 PixelAspectRatio;
    if (aspect_ratio_idc<Avc_PixelAspectRatio_Size)
        PixelAspectRatio=Avc_PixelAspectRatio[aspect_ratio_idc];
    else if (sar_height>0)
        PixelAspectRatio=((float)sar_width)/sar_height;
    else
        PixelAspectRatio=1; //Unknown

    Stream_Prepare(Stream_General);
    Fill("Format", "AVC");
    Stream_Prepare(Stream_Video);
    Fill("Codec", "AVC");

    Fill("Codec_Profile", ProfileS+_T("@L")+LevelS);
    Fill("Width", Width);
    Fill("Height", Height);
    Fill("PixelAspectRatio", PixelAspectRatio);
    if (Height!=0)
        Fill("DisplayAspectRatio", ((float)Width)/Height*PixelAspectRatio);

    if (File_Offset+Buffer_Size<File_Size)
    {
        NextCode_Clear();

        Info("AVC, Jumping to end of file");
        File_Offset=File_Size;
    }
}

//---------------------------------------------------------------------------
// Packet "06"
void File_Avc::sei()
{
    Element_Name("sei");

    //Parsing
    int32u payloadType=0, payloadSize=0;
    int8u payload_type_byte, payload_size_byte;
    Element_Begin("sei header");
        do
        {
            Get_B1 (payload_type_byte,                          "payload_type_byte");
            payloadType+=payload_type_byte;
        }
        while(payload_type_byte==0xFF);
        do
        {
            Get_B1 (payload_size_byte,                          "payload_size_byte");
            payloadSize+=payload_size_byte;
        }
        while(payload_size_byte==0xFF);
    Element_End();

    switch (payloadType)
    {
        case  0 :   sei_buffering_period(payloadSize); break;
        case  1 :   sei_pic_timing(payloadSize); break;
        case  5 :   sei_user_data_unregistered(payloadSize); break;
        default :
                    Element_Info("unknown");
                    Skip_XX(payloadSize,                        "data");
    }
}

//---------------------------------------------------------------------------
void File_Avc::sei_buffering_period(int32u payloadSize)
{
    Element_Info("buffering_period");

    //Parsing
    BS_Begin();
    Skip_UE(                                                    "seq_parameter_set_id");
    //TODO...
}

//---------------------------------------------------------------------------
void File_Avc::sei_pic_timing(int32u payloadSize)
{
    Element_Info("pic_timing");

    //Parsing
    Skip_XX(payloadSize,                                        "data");
    //TODO...
}

//---------------------------------------------------------------------------
void File_Avc::sei_user_data_unregistered(int32u payloadSize)
{
    Element_Info("user_data_unregistered");

    //Parsing
    int128u uuid_iso_iec_11578;
    Get_B16(uuid_iso_iec_11578,                                 "uuid_iso_iec_11578");

    switch (uuid_iso_iec_11578.hi)
    {
        case  0xDC45E9BDE6D948B7LL : Element_Info("x264");
                                     sei_user_data_unregistered_x264(payloadSize-16); break;
        case  0xFB574A60AC924E68LL : Element_Info("eavc");
                                     sei_user_data_unregistered_x264(payloadSize-16); break;
        default :
                    Element_Info("unknown");
                    Skip_XX(payloadSize-8,                      "data");
    }
}

//---------------------------------------------------------------------------
void File_Avc::sei_user_data_unregistered_x264(int32u payloadSize)
{
    //Parsing
    Ztring Data;
    Peek_Local(payloadSize, Data);
    size_t Data_Pos;
    size_t Data_Pos_Before=0;
    do
    {
        Data_Pos=Data.find(_T(" - "), Data_Pos_Before);
        if (Data_Pos==std::string::npos)
            Data_Pos=Data.size();
        if (Data.find(_T("options: "), Data_Pos_Before)==Data_Pos_Before)
        {
            Element_Begin("options");
            size_t Options_Pos;
            size_t Options_Pos_Before=Data_Pos_Before;
            do
            {
                Options_Pos=Data.find(_T(" "), Options_Pos_Before);
                if (Options_Pos==std::string::npos)
                    Options_Pos=Data.size();
                Skip_Local(Options_Pos-Options_Pos_Before,      "option");
                Options_Pos_Before=Options_Pos;
                if (Options_Pos_Before+3<=Data.size())
                {
                    Skip_Local(1,                               "separator");
                    Options_Pos_Before+=1;
                }
            }
            while (Options_Pos_Before!=Data.size());
            Element_End();
        }
        else
            Skip_Local(Data_Pos-Data_Pos_Before,                "data");
        Data_Pos_Before=Data_Pos;
        if (Data_Pos_Before+3<=Data.size())
        {
            Skip_Local(3,                                       "separator");
            Data_Pos_Before+=3;
        }
    }
    while (Data_Pos_Before!=Data.size());
}

//---------------------------------------------------------------------------
// Packet "07"
void File_Avc::seq_parameter_set()
{
    Element_Name("seq_parameter_set");

    int32u pic_order_cnt_type;
    bool frame_cropping_flag, vui_parameters_present_flag;
    Get_B1 (profile_idc,                                        "profile_idc");
    BS_Begin();
    Skip_SB(                                                    "constraint_set0_flag");
    Skip_SB(                                                    "constraint_set1_flag");
    Skip_SB(                                                    "constraint_set2_flag");
    Skip_SB(                                                    "constraint_set3_flag");
    Skip_SB(                                                    "reserved_zero_4bits");
    Skip_SB(                                                    "reserved_zero_4bits");
    Skip_SB(                                                    "reserved_zero_4bits");
    Skip_SB(                                                    "reserved_zero_4bits");
    Get_S1 ( 8, level_idc,                                      "level_idc");
    Skip_UE(                                                    "seq_parameter_set_id");
    if (profile_idc==0x64
     || profile_idc==0x6E
     || profile_idc==0x7A
     || profile_idc==0x90) //High profiles
    {
        int32u chroma_format_idc;
        bool seq_scaling_matrix_present_flag;
        Get_UE (chroma_format_idc,                              "chroma_format_idc");
        if (chroma_format_idc==3)
            Skip_UE (                                           "residual_colour_transform_flag");
        Skip_UE(                                                "bit_depth_luma_minus8");
        Skip_UE(                                                "bit_depth_chroma_minus8");
        Skip_SB(                                                "qpprime_y_zero_transform_bypass_flag");
        Get_SB (seq_scaling_matrix_present_flag,                "seq_scaling_matrix_present_flag");
        if (seq_scaling_matrix_present_flag)
        {
            bool Temp;
            for (int32u Pos=0; Pos<8; Pos++)
            {
                Get_SB (Temp,                                   "seq_scaling_list_present_flag");
                if (Temp)
                    scaling_list(Pos<6?16:64); //Never tested!
            }
        }
    }
    Get_UE (log2_max_frame_num_minus4,                          "log2_max_frame_num_minus4");
    Get_UE (pic_order_cnt_type,                                 "pic_order_cnt_type");
    if (pic_order_cnt_type==0)
        Skip_UE(                                                "log2_max_pic_order_cnt_lsb_minus4");
    else if (pic_order_cnt_type==1)
    {
        int32u num_ref_frames_in_pic_order_cnt_cycle;
        Skip_SB(                                                "delta_pic_order_always_zero_flag");
        Skip_SE(                                                "offset_for_non_ref_pic");
        Skip_SE(                                                "offset_for_top_to_bottom_field");
        Get_UE (num_ref_frames_in_pic_order_cnt_cycle,          "num_ref_frames_in_pic_order_cnt_cycle");
        for(int32u Pos=0; Pos<num_ref_frames_in_pic_order_cnt_cycle; Pos++)
            Skip_SE(                                            "offset_for_ref_frame");
    }
    Skip_UE(                                                    "num_ref_frames");
    Skip_SB(                                                    "gaps_in_frame_num_value_allowed_flag");
    Get_UE (pic_width_in_mbs_minus1,                            "pic_width_in_mbs_minus1");
    Get_UE (pic_height_in_map_units_minus1,                     "pic_height_in_map_units_minus1");
    Get_SB (frame_mbs_only_flag,                                "frame_mbs_only_flag");
    if (!frame_mbs_only_flag)
        Skip_SB(                                                "mb_adaptive_frame_field_flag");
    Skip_SB(                                                    "direct_8x8_inference_flag");
    Get_SB (frame_cropping_flag,                                "frame_cropping_flag");
    if (frame_cropping_flag)
    {
        Skip_UE(                                                "frame_crop_left_offset");
        Skip_UE(                                                "frame_crop_right_offset");
        Skip_UE(                                                "frame_crop_top_offset");
        Skip_UE(                                                "frame_crop_bottom_offset");
    }
    Get_SB (vui_parameters_present_flag,                        "vui_parameters_present_flag");
    if (vui_parameters_present_flag)
        vui_parameters();

    FILLING_BEGIN();
        //NextCode
        NextCode_Clear();
        NextCode_Add(0x08);

        //Autorisation of other streams
        Stream[0x06].Searching_Payload=true; //sei
        for (int8u Pos=0x08; Pos<=0x1F; Pos++)
            Stream[Pos].Searching_Payload=true; //pic_parameter_set, access_unit_delimiter, end_of_seq, end_of_stream, filler_data, reserved
    FILLING_END();
}

//---------------------------------------------------------------------------
// Packet "08"
void File_Avc::pic_parameter_set()
{
    Element_Name("pic_parameter_set");

    //Parsing
    int32u num_slice_groups_minus1;
    BS_Begin();
    Skip_UE(                                                    "pic_parameter_set_id");
    Skip_UE(                                                    "seq_parameter_set_id");
    Skip_SB(                                                    "entropy_coding_mode_flag");
    Skip_SB(                                                    "pic_order_present_flag");
    Get_UE (num_slice_groups_minus1,                            "num_slice_groups_minus1");
    if (num_slice_groups_minus1>0)
    {
        int32u slice_group_map_type;
        Get_UE (slice_group_map_type,                           "slice_group_map_type");
        if (slice_group_map_type==0)
        {
            for (int32u Pos=0; Pos<=num_slice_groups_minus1; Pos++)
                Skip_UE(                                        "run_length_minus1");
        }
        else if (slice_group_map_type==2)
        {
            for (int32u Pos=0; Pos<num_slice_groups_minus1; Pos++)
            {
                Skip_UE(                                        "top_left");
                Skip_UE(                                        "bottom_right");
            }
        }
        else if (slice_group_map_type==3
              || slice_group_map_type==4
              || slice_group_map_type==5)
        {
            Skip_SB(                                            "slice_group_change_direction_flag");
            Skip_UE(                                            "slice_group_change_rate_minus1");
        }
        else if (slice_group_map_type==6)
        {
            int32u pic_size_in_map_units_minus1;
            Get_UE (pic_size_in_map_units_minus1,               "pic_size_in_map_units_minus1");
            int32u slice_group_id_Size=(int32u)(ceil(log((float32)(num_slice_groups_minus1+1))/log((float32)2))); //this is log2
            for (int32u Pos=0; Pos<=pic_size_in_map_units_minus1; Pos++)
                Skip_S4(slice_group_id_Size,                    "slice_group_id");
        }
    }
    Skip_UE(                                                    "num_ref_idx_l0_active_minus1");
    Skip_UE(                                                    "num_ref_idx_l1_active_minus1");
    Skip_SB(                                                    "weighted_pred_flag");
    Skip_S1(2,                                                  "weighted_bipred_idc");
    Skip_SE(                                                    "pic_init_qp_minus26");
    Skip_SE(                                                    "pic_init_qs_minus26");
    Skip_SE(                                                    "chroma_qp_index_offset");
    Skip_SB(                                                    "deblocking_filter_control_present_flag");
    Skip_SB(                                                    "constrained_intra_pred_flag");
    Skip_SB(                                                    "redundant_pic_cnt_present_flag");

    FILLING_BEGIN();
        //NextCode
        NextCode_Clear();
        NextCode_Add(0x05);
        NextCode_Add(0x06);

        //Autorisation of other streams
        for (int8u Pos=0x01; Pos<=0x06; Pos++)
            Stream[Pos].Searching_Payload=true; //Coded slice...
    FILLING_END();
}

//---------------------------------------------------------------------------
// Packet "09"
void File_Avc::access_unit_delimiter()
{
    Element_Name("access_unit_delimiter");

    int8u primary_pic_type;
    BS_Begin();
    Get_S1 ( 3, primary_pic_type,                               "primary_pic_type"); Param_Info(Avc_primary_pic_type[primary_pic_type]);
    BS_End();
}

//***************************************************************************
// SubElements
//***************************************************************************

//---------------------------------------------------------------------------
void File_Avc::scaling_list(int32u ScalingList_Size)
{
    //From http://mpeg4ip.cvs.sourceforge.net/mpeg4ip/mpeg4ip/util/h264/main.cpp?revision=1.17&view=markup, never tested
    double Scale_Delta;
    int32u Scale_Delta_NeverTested, Scale_Last=8, Scale_Next=8;
    for (int32u Pos=0; Pos<ScalingList_Size; Pos++)
    {
        if (Scale_Next!=0)
        {
            Get_UE (Scale_Delta_NeverTested,                    "Scale_Delta_NeverTested");
            Scale_Delta=Scale_Delta_NeverTested;
            Scale_Next=(Scale_Last+(int32u)(Scale_Delta)+256)%256;
            Scale_Last=Scale_Next;
        }
    }
}

//---------------------------------------------------------------------------
void File_Avc::vui_parameters()
{
    TEST_SB_SKIP(                                               "aspect_ratio_info_present_flag");
        Get_S1 (8, aspect_ratio_idc,                            "aspect_ratio_idc");
        if (aspect_ratio_idc==0xFF)
        {
            Get_S2 (16, sar_width,                              "sar_width");
            Get_S2 (16, sar_height,                             "sar_height");
        }
    TEST_SB_END();
}

//***************************************************************************
// Helpers
//***************************************************************************

//---------------------------------------------------------------------------
bool File_Avc::Synchronize()
{
    //Synchronizing
    while (Buffer_Offset+4<=Buffer_Size)
    {
         while (Buffer_Offset+4<=Buffer_Size
          && !(CC3(Buffer+Buffer_Offset  )==0x000001
            && CC1(Buffer+Buffer_Offset+3)< 0x80)) //bit 7 can not be 1
                 Buffer_Offset++;

        if (Buffer_Offset+4<=Buffer_Size)//Testing if header is coherant
        {
            int8u nal_unit_type=CC1(Buffer+Buffer_Offset+3);
            bool nal_ref_idc=(nal_unit_type&0x60)?true:false;
            nal_unit_type&=0x1F;
            if (Stream[nal_unit_type].Searching_Payload
             && ((nal_ref_idc  && (nal_unit_type==0x05 || nal_unit_type==0x07 || nal_unit_type==0x08))
              || (!nal_ref_idc && (nal_unit_type==0x06 || nal_unit_type==0x09 || nal_unit_type==0x10 || nal_unit_type==0x0A || nal_unit_type==0x0B))))
                 break;
            else
                Buffer_Offset++;
        }
    }
    if (Buffer_Offset+4>Buffer_Size)
    {
        //Parsing last bytes
        if (Buffer_Offset+4==Buffer_Size)
        {
            if (CC3(Buffer+Buffer_Offset)!=0x000001)
            {
                Buffer_Offset++;
                if (CC3(Buffer+Buffer_Offset)!=0x000001)
                {
                    Buffer_Offset++;
                    if (CC2(Buffer+Buffer_Offset)!=0x0000)
                    {
                        Buffer_Offset++;
                        if (CC1(Buffer+Buffer_Offset)!=0x00)
                            Buffer_Offset++;
                    }
                }
            }
        }

        return false;
    }

    //Synched is OK
    Synched=true;
    return true;
}

//---------------------------------------------------------------------------
bool File_Avc::Header_Parser_QuickSearch()
{
    while (           Buffer_Offset+4<=Buffer_Size
      &&   CC3(Buffer+Buffer_Offset)==0x000001)
    {
        //Getting start_code
        int8u start_code=CC1(Buffer+Buffer_Offset+3)&0x1F;

        //Searching start
        if (Stream[start_code].Searching_Payload)
            return true;

        //Getting size
        Buffer_Offset+=4;
        while(Buffer_Offset+4<=Buffer_Size && CC3(Buffer+Buffer_Offset)!=0x000001)
            Buffer_Offset++;
    }

    if (Buffer_Offset+4>Buffer_Size)
    {
        Synched=false;
        Synchronize();
        return false;
    }
    return true;
}

//---------------------------------------------------------------------------
bool File_Avc::Detect_NonAVC ()
{
    //File_Size
    if (File_Size<188*4)
        return false; //We can't do detection

    //Element_Size
    if (Buffer_Size<188*4)
        return true; //Must wait for more data

    //Detect mainly DAT files, and the parser is not enough precise to detect them later
    if (CC4(Buffer)==CC4("RIFF"))
    {
        Finnished();
        return true;
    }

    //Detect TS files, and the parser is not enough precise to detect them later
    while (Buffer_Offset<188 && CC1(Buffer+Buffer_Offset)!=0x47) //Look for first Sync word
        Buffer_Offset++;
    if (Buffer_Offset<188 && CC1(Buffer+Buffer_Offset+188)==0x47 && CC1(Buffer+Buffer_Offset+188*2)==0x47 && CC1(Buffer+Buffer_Offset+188*3)==0x47)
    {
        Finnished();
        return true;
    }
    Buffer_Offset=0;

    //Seems OK
    return false;
}

//***************************************************************************
// Information
//***************************************************************************

//---------------------------------------------------------------------------
void File_Avc::HowTo(stream_t StreamKind)
{
    switch (StreamKind)
    {
        case (Stream_General) :
            Fill_HowTo("Format", "R");
            break;
        case (Stream_Video) :
            Fill_HowTo("Codec", "R");
            break;
        case (Stream_Audio) :
            break;
        case (Stream_Text) :
            break;
        case (Stream_Chapters) :
            break;
        case (Stream_Image) :
            break;
        case (Stream_Menu) :
            break;
        case (Stream_Max) :
            break;
    }
}

} //NameSpace

#endif //MEDIAINFO_AVC_YES
